@@ -40,6 +40,9 @@ class Base:
     def __init__(self, sock=None, connection_type=None, receive_bytes=4096,
                  ping_interval=None, max_message_size=None,
                  thread_class=None, event_class=None, selector_class=None):
+        #: The name of the subprotocol chosen for the WebSocket connection.
+        self.subprotocol = None
+
         self.sock = sock
         self.receive_bytes = receive_bytes
         self.ping_interval = ping_interval
@@ -47,6 +50,7 @@ class Base:
         self.pong_received = True
         self.input_buffer = []
         self.incoming_message = None
+        self.incoming_message_len = 0
         self.connected = False
         self.is_server = (connection_type == ConnectionType.SERVER)
         self.close_reason = CloseReason.NO_STATUS_RCVD
@@ -59,11 +63,9 @@ class Base:
             import threading
             event_class = threading.Event
         if selector_class is None:
-            import selectors
             selector_class = selectors.DefaultSelector
-
+        self.selector_class = selector_class
         self.event = event_class()
-        self.selector = selector_class()
 
         self.ws = WSConnection(connection_type)
         self.handshake()
@@ -96,7 +98,7 @@ class Base:
         """Receive data over the WebSocket connection.
 
         :param timeout: Amount of time to wait for the data, in seconds. Set
-                        to ``None`` (the default) to wait undefinitely. Set
+                        to ``None`` (the default) to wait indefinitely. Set
                         to 0 to read without blocking.
 
         The data received is returned, as ``bytes`` or ``str``, depending on
@@ -114,7 +116,7 @@ class Base:
         """Close the WebSocket connection.
 
         :param reason: A numeric status code indicating the reason of the
-                       closure, as defined by the WebSocket specifiation. The
+                       closure, as defined by the WebSocket specification. The
                        default is 1000 (normal closure).
         :param message: A text message to be sent to the other side.
         """
@@ -128,18 +130,24 @@ class Base:
             pass
         self.connected = False
 
+    def choose_subprotocol(self, request):  # pragma: no cover
+        # The method should return the subprotocol to use, or ``None`` if no
+        # subprotocol is chosen. Can be overridden by subclasses that implement
+        # the server-side of the WebSocket protocol.
+        return None
+
     def _thread(self):
         sel = None
         if self.ping_interval:
             next_ping = time() + self.ping_interval
-            sel = self.selector
+            sel = self.selector_class()
             sel.register(self.sock, selectors.EVENT_READ, True)
 
         while self.connected:
             try:
                 if sel:
                     now = time()
-                    if not sel.select(next_ping - now):
+                    if next_ping <= now or not sel.select(next_ping - now):
                         # we reached the timeout, we have to send a ping
                         if not self.pong_received:
                             self.close(reason=CloseReason.POLICY_VIOLATION,
@@ -147,7 +155,7 @@ class Base:
                             break
                         self.pong_received = False
                         self.sock.send(self.ws.send(Ping()))
-                        next_ping += self.ping_interval
+                        next_ping = max(now, next_ping) + self.ping_interval
                         continue
                 in_data = self.sock.recv(self.receive_bytes)
                 if len(in_data) == 0:
@@ -156,8 +164,11 @@ class Base:
                 self.connected = False
                 self.event.set()
                 break
+
             self.ws.receive_data(in_data)
             self.connected = self._handle_events()
+        sel.close() if sel else None
+        self.sock.close()
 
     def _handle_events(self):
         keep_going = True
@@ -165,13 +176,16 @@ class Base:
         for event in self.ws.events():
             try:
                 if isinstance(event, Request):
+                    self.subprotocol = self.choose_subprotocol(event)
                     out_data += self.ws.send(AcceptConnection(
+                        subprotocol=self.subprotocol,
                         extensions=[PerMessageDeflate()]))
                 elif isinstance(event, CloseConnection):
                     if self.is_server:
                         out_data += self.ws.send(event.response())
                     self.close_reason = event.code
                     self.close_message = event.reason
+                    self.connected = False
                     self.event.set()
                     keep_going = False
                 elif isinstance(event, Ping):
@@ -179,21 +193,51 @@ class Base:
                 elif isinstance(event, Pong):
                     self.pong_received = True
                 elif isinstance(event, (TextMessage, BytesMessage)):
+                    self.incoming_message_len += len(event.data)
                     if self.max_message_size and \
-                            len(event.data) > self.max_message_size:
+                            self.incoming_message_len > self.max_message_size:
                         out_data += self.ws.send(CloseConnection(
                             CloseReason.MESSAGE_TOO_BIG, 'Message is too big'))
                         self.event.set()
                         keep_going = False
                         break
                     if self.incoming_message is None:
+                        # store message as is first
+                        # if it is the first of a group, the message will be
+                        # converted to bytearray on arrival of the second
+                        # part, since bytearrays are mutable and can be
+                        # concatenated more efficiently
                         self.incoming_message = event.data
+                    elif isinstance(event, TextMessage):
+                        if not isinstance(self.incoming_message, bytearray):
+                            # convert to bytearray and append
+                            self.incoming_message = bytearray(
+                                (self.incoming_message + event.data).encode())
+                        else:
+                            # append to bytearray
+                            self.incoming_message += event.data.encode()
                     else:
-                        self.incoming_message += event.data
+                        if not isinstance(self.incoming_message, bytearray):
+                            # convert to mutable bytearray and append
+                            self.incoming_message = bytearray(
+                                self.incoming_message + event.data)
+                        else:
+                            # append to bytearray
+                            self.incoming_message += event.data
                     if not event.message_finished:
                         continue
-                    self.input_buffer.append(self.incoming_message)
+                    if isinstance(self.incoming_message, (str, bytes)):
+                        # single part message
+                        self.input_buffer.append(self.incoming_message)
+                    elif isinstance(event, TextMessage):
+                        # convert multi-part message back to text
+                        self.input_buffer.append(
+                            self.incoming_message.decode())
+                    else:
+                        # convert multi-part message back to bytes
+                        self.input_buffer.append(bytes(self.incoming_message))
                     self.incoming_message = None
+                    self.incoming_message_len = 0
                     self.event.set()
                 else:  # pragma: no cover
                     pass
@@ -217,6 +261,8 @@ class Server(Base):
                     does this in its own different way. Werkzeug, Gunicorn,
                     Eventlet and Gevent are the only web servers that are
                     currently supported.
+    :param subprotocols: A list of supported subprotocols, or ``None`` (the
+                         default) to disable subprotocol negotiation.
     :param receive_bytes: The size of the receive buffer, in bytes. The
                           default is 4096.
     :param ping_interval: Send ping packets to clients at the requested
@@ -239,10 +285,13 @@ class Server(Base):
                            ``selectors.DefaultSelector`` class from the Python
                            standard library.
     """
-    def __init__(self, environ, receive_bytes=4096, ping_interval=None,
-                 max_message_size=None, thread_class=None, event_class=None,
-                 selector_class=None):
+    def __init__(self, environ, subprotocols=None, receive_bytes=4096,
+                 ping_interval=None, max_message_size=None, thread_class=None,
+                 event_class=None, selector_class=None):
         self.environ = environ
+        self.subprotocols = subprotocols or []
+        if isinstance(self.subprotocols, str):
+            self.subprotocols = [self.subprotocols]
         self.mode = 'unknown'
         sock = None
         if 'werkzeug.socket' in environ:
@@ -285,12 +334,36 @@ class Server(Base):
         self.ws.receive_data(in_data)
         self.connected = self._handle_events()
 
+    def choose_subprotocol(self, request):
+        """Choose a subprotocol to use for the WebSocket connection.
+
+        The default implementation selects the first protocol requested by the
+        client that is accepted by the server. Subclasses can override this
+        method to implement a different subprotocol negotiation algorithm.
+
+        :param request: A ``Request`` object.
+
+        The method should return the subprotocol to use, or ``None`` if no
+        subprotocol is chosen.
+        """
+        for subprotocol in request.subprotocols:
+            if subprotocol in self.subprotocols:
+                return subprotocol
+        return None
+
 
 class Client(Base):
     """This class implements a WebSocket client.
 
     :param url: The connection URL. Both ``ws://`` and ``wss://`` URLs are
                 accepted.
+    :param subprotocols: The name of the subprotocol to use, or a list of
+                         subprotocol names in order of preference. Set to
+                         ``None`` (the default) to not use a subprotocol.
+    :param headers: A dictionary or list of tuples with additional HTTP headers
+                    to send with the connection request. Note that custom
+                    headers are not supported by the WebSocket protocol, so the
+                    use of this parameter is not recommended.
     :param receive_bytes: The size of the receive buffer, in bytes. The
                           default is 4096.
     :param ping_interval: Send ping packets to the server at the requested
@@ -313,9 +386,10 @@ class Client(Base):
                         objects. The default is the `threading.Event`` class
                         from the Python standard library.
     """
-    def __init__(self, url, receive_bytes=4096, ping_interval=None,
-                 max_message_size=None, ssl_context=None, thread_class=None,
-                 event_class=None, selector_class=None):
+    def __init__(self, url, subprotocols=None, headers=None,
+                 receive_bytes=4096, ping_interval=None, max_message_size=None,
+                 ssl_context=None, thread_class=None, event_class=None,
+                 selector_class=None):
         parsed_url = urlsplit(url)
         is_secure = parsed_url.scheme in ['https', 'wss']
         self.host = parsed_url.hostname
@@ -323,9 +397,19 @@ class Client(Base):
         self.path = parsed_url.path
         if parsed_url.query:
             self.path += '?' + parsed_url.query
+        self.subprotocols = subprotocols or []
+        if isinstance(self.subprotocols, str):
+            self.subprotocols = [self.subprotocols]
+
+        self.extra_headeers = []
+        if isinstance(headers, dict):
+            for key, value in headers.items():
+                self.extra_headeers.append((key, value))
+        elif isinstance(headers, list):
+            self.extra_headeers = headers
 
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        if is_secure:
+        if is_secure:  # pragma: no cover
             if ssl_context is None:
                 ssl_context = ssl.create_default_context(
                     purpose=ssl.Purpose.SERVER_AUTH)
@@ -338,16 +422,25 @@ class Client(Base):
                          thread_class=thread_class, event_class=event_class)
 
     def handshake(self):
-        out_data = self.ws.send(Request(host=self.host, target=self.path))
+        out_data = self.ws.send(Request(host=self.host, target=self.path,
+                                        subprotocols=self.subprotocols,
+                                        extra_headers=self.extra_headeers))
         self.sock.send(out_data)
 
-        in_data = self.sock.recv(self.receive_bytes)
-        self.ws.receive_data(in_data)
-        event = next(self.ws.events())
+        while True:
+            in_data = self.sock.recv(self.receive_bytes)
+            self.ws.receive_data(in_data)
+            try:
+                event = next(self.ws.events())
+            except StopIteration:  # pragma: no cover
+                pass
+            else:  # pragma: no cover
+                break
         if isinstance(event, RejectConnection):  # pragma: no cover
             raise ConnectionError(event.status_code)
         elif not isinstance(event, AcceptConnection):  # pragma: no cover
             raise ConnectionError(400)
+        self.subprotocol = event.subprotocol
         self.connected = True
 
     def close(self, reason=None, message=None):
